/*
  Copyright 2008 William Poetra Yoga Hadisoeseno <williampoetra@gmail.com>
  This file is licensed under GPL v2.

  !ATTENTION!
  Contains code stolen from jz4740_usbtool.c (GPL v2) by Maurus Cuelenaere
                                                         (pls dun tell)
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <libgen.h>

#include <usb.h>

void usage()
{
  fprintf(stderr, "usage: onda_restore vx747 1.bin 2.bin 3.bin 4.bin 5.bin\n");
  fprintf(stderr, "       onda_restore vx767 1.bin 2.bin 3.bin 4.bin 5.bin VX767_V1.0.dl\n");
  exit(1);
}

uint32_t sum1(char *data, int len)
{
  uint32_t sum;
  int i;

  sum = 0;
  for (i = 0; i < len; ++i)
    sum += (unsigned char) data[i];

  return sum;
}

#define BLOCK_SIZE 2048

int get_filesize(char *file)
{
  struct stat statbuf;
  stat(file, &statbuf);
  return statbuf.st_size;
}

void read_file(char *buf, int len, char *file)
{
  int fd, i, n, r;

  n = len / BLOCK_SIZE;
  r = len % BLOCK_SIZE;

  fd = open(file, O_RDONLY);

  for (i = 0; i < n; ++i)
    read(fd, buf + i * BLOCK_SIZE, BLOCK_SIZE);

  if (r > 0)
    read(fd, buf + n * BLOCK_SIZE, r);

  close(fd);
}

typedef struct {
  char *name;
  char *data;
  int len;
} bin_t;

#define USB_VEN 0x601A
#define USB_DEV 0x4740
#define EP_BULK_TO 0x01
#define TIMEOUT 5000
usb_dev_handle *device;

enum USB_JZ4740_REQUEST
{
    VR_GET_CPU_INFO = 0,
    VR_SET_DATA_ADDRESS,
    VR_SET_DATA_LENGTH,
    VR_FLUSH_CACHES,
    VR_PROGRAM_START1,
    VR_PROGRAM_START2,
    VR_NOR_OPS,
    VR_NAND_OPS,
    VR_SDRAM_OPS,
    VR_CONFIGURATION
};

void usb_dev_open()
{
  struct usb_bus *bus;
  struct usb_device *dev;

  printf("USB initialization...");

  usb_init();
  usb_find_busses();
  usb_find_devices();

  for (bus = usb_get_busses(); bus != NULL; bus = bus->next)
    for (dev = bus->devices; dev != NULL; dev = dev->next)
      if (dev->descriptor.idVendor == USB_VEN
          && dev->descriptor.idProduct == USB_DEV)
        goto found;

  printf("\nNo device found, exiting.\n");
  exit(1);

found:
  printf(" Device found.\n");
  device = usb_open(dev);
  usb_claim_interface(device, 0);
}

void usb_dev_close()
{
  printf("Releasing interface...");

  usb_release_interface(device, 0);

  printf(" OK\n");
}

int send_cmd(int cmd, uint32_t arg)
{
  return usb_control_msg(device, USB_ENDPOINT_OUT | USB_TYPE_VENDOR,
                         cmd, arg >> 16, arg & 0xffff,
                         NULL, 0, TIMEOUT);
}

void get_cpu()
{
  char data[9];
  int req_in_ven = USB_ENDPOINT_IN | USB_TYPE_VENDOR;
  int len;

  //printf("GET CPU");

  len = usb_control_msg(device, req_in_ven, VR_GET_CPU_INFO, 0, 0, data, 8, TIMEOUT);
  if (len < 0) {
    printf("\nError trying to get CPU model, exiting.\n");
    usb_dev_close();
    exit(1);
  }

  memset(data + len, 0, 9 - len);
  //printf(", got: %s\n", data);
}

void jz_run1(uint32_t addr)
{
  send_cmd(VR_PROGRAM_START1, addr);
}

void jz_run2(uint32_t addr)
{
  send_cmd(VR_PROGRAM_START2, addr);
}

void flush_dev()
{
  int req_out_ven = USB_ENDPOINT_OUT | USB_TYPE_VENDOR;
  get_cpu();
  send_cmd(VR_FLUSH_CACHES, 0);
  get_cpu();
}

void get_data(char *buf, int len, uint32_t addr)
{
  get_cpu();
  send_cmd(VR_SET_DATA_ADDRESS, addr);
  send_cmd(VR_SET_DATA_LENGTH, len);
  usb_bulk_read(device, USB_ENDPOINT_IN | EP_BULK_TO, buf, len, TIMEOUT);
}

void put_data(char *buf, int len, uint32_t addr)
{
  get_cpu();
  send_cmd(VR_SET_DATA_ADDRESS, addr);
  usb_bulk_write(device, USB_ENDPOINT_OUT | EP_BULK_TO, buf, len, TIMEOUT);
}

void send_data(char *buf, int len, uint32_t addr)
{
  char *mybuf;

  printf("Sending data...");
  put_data(buf, len, addr);
  printf(" OK\n");

  printf("Rereading data...");
  mybuf = malloc(len);
  get_data(mybuf, len, addr);
  printf(" OK\n");

  printf("Comparing...");
  if (!memcmp(buf, mybuf, len)) {
    printf(" OK\n");
  } else {
    printf(" ERROR: sent data is not the same as retrieved data\n");
    printf("Exiting...\n");
    usb_dev_close();
  }
}

void restore_vx747(char *file_1, char *file_2, char *file_3, char *file_4, char *file_5)
{
  int len;
  bin_t bin1, bin2, bin3, bin4, bin5;
  uint32_t ret1, ret2;
  uint8_t ret3;
  uint32_t arg1[2], arg2;

  // prepare the data

  printf("Preparing data...");

  len = get_filesize(file_1);
  bin1.name = basename(strdup(file_1));
  bin1.data = malloc(len);
  bin1.len = len;
  read_file(bin1.data, len, file_1);

  len = get_filesize(file_2);
  bin2.name = basename(strdup(file_2));
  bin2.data = malloc(len);
  bin2.len = len;
  read_file(bin2.data, len, file_2);

  len = get_filesize(file_3);
  bin3.name = basename(strdup(file_3));
  bin3.data = malloc(len);
  bin3.len = len;
  read_file(bin3.data, len, file_3);

  len = get_filesize(file_4);
  bin4.name = basename(strdup(file_4));
  bin4.data = malloc(len);
  bin4.len = len;
  read_file(bin4.data, len, file_4);

  len = get_filesize(file_5);
  bin5.name = basename(strdup(file_5));
  bin5.data = malloc(len);
  bin5.len = len;
  read_file(bin5.data, len, file_5);

  arg1[0] = sum1(bin2.data, bin2.len);
  arg1[1] = bin2.len;

  arg2 = 1;

  printf(" OK\n");

  // prepare

  usb_dev_open();

  // part 1

  printf("Part 1\n");

  send_data(bin1.data, bin1.len, 0x80000000);
  jz_run1(0x80000000);
  sleep(3);
  get_data((char *) &ret1, 4, 0xb3020060);
  if (ret1 != 0)
    printf("WARNING: retval is not 0\n");
  get_cpu();

  // part 2

  printf("Part 2\n");

  flush_dev();
  send_data(bin2.data, bin2.len, 0x80000000);
  flush_dev();
  send_data((char *) arg1, 8, 0x80d00000);
  flush_dev();
  send_data(bin3.data, bin3.len, 0x80e00000);
  flush_dev();
  send_data((char *) &arg2, 4, 0x80004000);
  flush_dev();
  send_data(bin4.data, bin4.len, 0x80fd0000);
  flush_dev();
  jz_run2(0x80fd0004);
  get_data((char *) &ret2, 4, 0x80004004);
  if (ret2 != 1)
    printf("WARNING: retval is not 1\n");
  get_data(&ret3, 1, 0x80004008);
  if (ret3 != 1)
    printf("WARNING: retval is not 1\n");
  sleep(2);

  // part 3

  printf("Part 3\n");

  send_data(bin5.data, bin5.len, 0x80e00000);
  flush_dev();
  jz_run2(0x80e00008);

  // clean-up

  usb_dev_close();
}

void restore_vx767(char *file_1, char *file_2, char *file_3, char *file_4, char *file_5, char *file_vx767_dl)
{
  int len;
  bin_t bin1, bin2, bin3, bin4, bin5;
  uint32_t ret1, ret2;
  uint8_t ret3;
  uint32_t arg1[2], arg2;
  uint32_t sum;

  // prepare the data

  printf("Preparing data...");

  len = get_filesize(file_1);
  bin1.name = basename(strdup(file_1));
  bin1.data = malloc(len);
  bin1.len = len;
  read_file(bin1.data, len, file_1);

  len = get_filesize(file_2);
  bin2.name = basename(strdup(file_2));
  bin2.data = malloc(len);
  bin2.len = len;
  read_file(bin2.data, len, file_2);

  len = get_filesize(file_3);
  bin3.name = basename(strdup(file_3));
  bin3.data = malloc(len + 65536);
  bin3.len = len + 65536;
  read_file(bin3.data + 65536, len, file_3);
  len = get_filesize(file_vx767_dl);
  read_file(bin3.data + 12, len, file_vx767_dl);
  sum = sum1(bin3.data + 12, len);
  memcpy(bin3.data, &bin3.len, 4);
  memcpy(bin3.data + 4, &sum, 4);
  memcpy(bin3.data + 8, &len, 4);

  len = get_filesize(file_4);
  bin4.name = basename(strdup(file_4));
  bin4.data = malloc(len);
  bin4.len = len;
  read_file(bin4.data, len, file_4);

  len = get_filesize(file_5);
  bin5.name = basename(strdup(file_5));
  bin5.data = malloc(len + 65536);
  bin5.len = len + 65536;
  read_file(bin5.data + 65536, len, file_5);
  len = get_filesize(file_vx767_dl);
  read_file(bin5.data + 12, len, file_vx767_dl);
  sum = sum1(bin5.data + 12, len);
  memcpy(bin5.data, &bin5.len, 4);
  memcpy(bin5.data + 4, &sum, 4);
  memcpy(bin5.data + 8, &len, 4);

  arg1[0] = sum1(bin2.data, bin2.len);
  arg1[1] = bin2.len;

  arg2 = 1;

  printf(" OK\n");

  // prepare

  usb_dev_open();

  // part 1

  printf("Part 1\n");

  send_data(bin1.data, bin1.len, 0x80000000);
  jz_run1(0x80000000);
  sleep(3);
  get_data((char *) &ret1, 4, 0xb3020060);
  if (ret1 != 0)
    printf("WARNING: retval is not 0\n");
  get_cpu();

  // part 2

  printf("Part 2\n");

  flush_dev();
  send_data(bin2.data, bin2.len, 0x80000000);
  flush_dev();
  send_data((char *) arg1, 8, 0x80d00000);
  flush_dev();
  send_data(bin3.data, bin3.len, 0x80e00000);
  flush_dev();
  send_data((char *) &arg2, 4, 0x80004000);
  flush_dev();
  send_data(bin4.data, bin4.len, 0x80fd0000);
  flush_dev();
  jz_run2(0x80fd0004);
  get_data((char *) &ret2, 4, 0x80004004);
  if (ret2 != 1)
    printf("WARNING: retval is not 1\n");
  get_data(&ret3, 1, 0x80004008);
  if (ret3 != 1)
    printf("WARNING: retval is not 1\n");
  sleep(2);

  // part 3

  printf("Part 3\n");

  send_data(bin5.data, bin5.len, 0x80e00000);
  flush_dev();
  jz_run2(0x80e10008);

  // clean-up

  usb_dev_close();
}


int main(int argc, char **argv)
{
  if (argc != 7 && argc != 8)
    usage();

  setvbuf(stdout, NULL, _IONBF, 0);

  if (!strcmp(argv[1], "vx747") && argc == 7)
    restore_vx747(argv[2], argv[3], argv[4], argv[5], argv[6]);
  else if (!strcmp(argv[1], "vx767") && argc == 8)
    restore_vx767(argv[2], argv[3], argv[4], argv[5], argv[6], argv[7]);
  else
    usage();

  return 0;
}

